/*
 * vdiff-seq.js - JavaScript-based visual differencing tool for sequential images
 * http://codh.rois.ac.jp/software/vdiffseqjs/
 *
 * Copyright 2022 Center for Open Data in the Humanities, Research Organization of Information and Systems
 * Released under the MIT license
 *
 * Core contributor: Jun HOMMA (@2SC1815J)
 *
 * Licenses of open source libraries, see acknowledgements.md
 */
var VDiffSeq = function(config) {
    'use strict';

    var $ = window.$;

    var APP_NAME = 'vdiff-seq.js';
    var VERSION  = '1.1.1+20230317';
    if (window.console) {
        console.log(APP_NAME + ' ' + VERSION); // eslint-disable-line no-console
    }
    var exportFunctions;

    //リテラルはさほど多くないので、i18n用のフレームワークは用いず、直接記述する。
    var lng = getLang(config);

    //IEはWebAssembly未対応（WebAssembly以外にも未対応のCSSなどがあるため動作対象外とする）
    var userAgent = window.navigator.userAgent.toLowerCase();
    if (userAgent.indexOf('msie') != -1 || userAgent.indexOf('trident') != -1) {
        showError((lng !== 'ja') ? 'Internet Explorer is not supported.' : 'Internet Explorerは未対応です。');
        return;
    } else if (typeof WebAssembly !== 'object') {
        showError((lng !== 'ja') ? 'WebAssembly is not supported by your browser.' : 'お使いのブラウザはWebAssembly未対応です。');
        return;
    }

    function getLang() {
        function getLangFromObject(obj) {
            if ($.isPlainObject(obj)) {
                if (obj.lang) { //表示言語指定
                    // lang: undefined などの場合は、ここでは決定しない。
                    if (obj.lang !== 'ja') {
                        return 'en'; //ja以外は全てenにfallback
                    } else {
                        return 'ja';
                    }
                }
            }
            return null;
        }
        //第1優先：GETパラメータ
        var params_ = getParams(location.search);
        var lang = getLangFromObject(params_);
        //第2優先：設定ファイル
        if (!lang) {
            lang = getLangFromObject(config);
        }
        //第3優先：ブラウザの言語設定
        if (!lang) {
            lang = (String(window.navigator.language || window.navigator.userLanguage || 'ja').substring(0, 2) !== 'ja' ? 'en' : 'ja');
        }
        return lang;
    }
    function getParams(search) {
        var query = search.substring(1);
        if (query !== '') {
            var params = query.split('&');
            var paramsObj = {};
            for (var i = 0; i < params.length; i++) {
                var elems = params[i].split('=');
                if (elems.length > 1) {
                    var key = decodeURIComponent(elems[0].replace(/\+/g, ' '));
                    var val = decodeURIComponent(elems[1].replace(/\+/g, ' '));
                    paramsObj[key] = val;
                }
            }
            return paramsObj;
        } else {
            return null;
        }
    }
    function showError(msg) {
        if ($.isPlainObject(config)) {
            if (config.id) {
                $('#' + config.id).text(msg);
            }
            if ($.isFunction(config.failCallback)) {
                config.failCallback();
            }
            if ($.isFunction(config.alwaysCallback)) {
                config.alwaysCallback();
            }
        }
    }

    function showComparedImages(config_, params_) {
        var config__ = $.isPlainObject(config_) ? config_ : {};
        var containerId = '#' + config__.id;
        var limitSize = config__.size;

        var params__ = $.isPlainObject(params_) ? params_ : {};
        //sequenceは、通常版vdiff.jsで用いる`data`オブジェクトの配列
        var sequence = $.isArray(params__.sequence) ? params__.sequence : [];
        var isGlobalCompareMode = (params__.mode); //true: 全体変化比較モード, false: 局所強調比較モード

        $(containerId).addClass('vdiffseqjs-container');

        if ($.fn.spin) {
            $('body').spin({ top: '50vh' });
        }

        //ブラウザのキャッシュから読み込まれると、crossOrigin: 'anonymous'が反映されずセキュリティエラーになるため再取得
        function getUrlWithTimeStamp(url, opt) {
            var url_ = url.toString();
            if (url_.toLowerCase().indexOf('data:') === 0 || url_.toLowerCase().indexOf('blob:') === 0 || opt.no_ts) {
                return url_;
            } else {
                return url_ + (url_.indexOf('?') !== -1 ? '&' : '?') + new Date().getTime();
            }
        }
        var $imgs = [];
        var imgUrlDict = [];
        for (var i = 0; i < sequence.length; i++) {
            if ($.isPlainObject(sequence[i])) {
                var imgUrl = sequence[i].img1;
                if (imgUrl && imgUrlDict.indexOf(imgUrl) < 0) {
                    $imgs.push($('<img>').attr({ src: getUrlWithTimeStamp(imgUrl, params__), crossOrigin: 'anonymous' }));
                    imgUrlDict.push(imgUrl);
                }
                imgUrl = sequence[i].img2;
                if (imgUrl && imgUrlDict.indexOf(imgUrl) < 0) {
                    $imgs.push($('<img>').attr({ src: getUrlWithTimeStamp(imgUrl, params__), crossOrigin: 'anonymous' }));
                    imgUrlDict.push(imgUrl);
                }
            }
        }
        var $imagesContainer = $('<div>').addClass('seq-image-container');
        for (i = 0; i < $imgs.length; i++) {
            $imagesContainer.append($('<div>').append($imgs[i]).css({ 'position': 'absolute', 'top': 0, 'left': 0, 'z-index': $imgs.length - i }));
        }
        $(containerId).append($imagesContainer);
        $imagesContainer.hide();

        //進捗表示の準備
        var $progressbarContainer = $('<div>').addClass('seq-progressbar-container');
        var $progressbar = $('<div>').addClass('seq-progressbar');
        var $progressLabel = $('<div>').addClass('seq-progress-label');
        $progressbar.append($progressLabel);
        $progressbarContainer.append($progressbar);
        $imagesContainer.append($progressbarContainer);

        //カルーセルの準備
        var $slickContainer = $('<ul>').addClass('slick-slider');
        $(containerId).append($slickContainer);

        //キャンバスなど
        var containerId_ = containerId.slice(1) + '-';
        var canvasPrefix = containerId_ + 'canvas';
        var $outputContainer = $('<div>').hide();
        var $canvasOutputArray = [];
        for (i = 0; i < $imgs.length - 1; i++) {
            var $canvasOutput_ = $('<canvas>').attr({ id: canvasPrefix + '_merged_highlighted_' + i});
            $canvasOutputArray.push($canvasOutput_);
            $outputContainer.append($canvasOutput_);
        }
        $(containerId).append($outputContainer);
        //累積的に射影変換を適用する前の変換結果を取っておく
        var $outputNoncumulativeContainer = $('<div>').hide();
        var $canvasOutputNoncumulativeArray = [];
        for (i = 0; i < $imgs.length - 1; i++) {
            var $canvasOutputNoncumulative_ = $('<canvas>').attr({ id: canvasPrefix + '_merged_highlighted_noncumulative_' + i});
            $canvasOutputNoncumulativeArray.push($canvasOutputNoncumulative_);
            $outputNoncumulativeContainer.append($canvasOutputNoncumulative_);
        }
        $(containerId).append($outputNoncumulativeContainer);

        // 比較結果の詳細表示用ボタン
        var detailButtonIcon = [
            '<i class="fas fa-angle-double-down"></i>',
            '<i class="fas fa-angle-double-up"></i>'
        ];
        var $detailButton = $('<button>').html(detailButtonIcon[0]).addClass('ui-button ui-widget ui-corner-all').attr({ type: 'button', title: (lng !== 'ja') ? 'Show detail' : '詳細表示' });
        var $detailButtonWrapper = $('<ul>').addClass('show-detail-wrapper');
        $detailButtonWrapper.append($detailButton);
        $detailButton.on('click', function() {
            $vdiffjs.toggle(0, function() {
                var visible = $(this).is(':visible') ? 1 : 0;
                $detailButton.html(detailButtonIcon[visible]);
                if (visible) {
                    var currentSlide = $slickContainer.slick('slickCurrentSlide');
                    showDetail(currentSlide);
                }
            });
        });

        // 比較結果の詳細表示用vdiff.js
        var vdiffjsId = containerId_ + '_vdiffjs_detail';
        var vdiffjsEditorId = containerId_ + '_vdiffjs_detail_editor';
        var $vdiffjs = $('<div>').attr({ id: vdiffjsId }).hide();
        var $vdiffjsEditor = $('<div>').attr({ id: vdiffjsEditorId }).hide();
        var $vdiffjsWrapper = $('<ul>').addClass('vdiffjs-detail-wrapper');
        $vdiffjsWrapper.append($('<li>').append($vdiffjs).append($vdiffjsEditor));
        $(containerId).append($vdiffjsWrapper);

        // カルーセルのスライド切り替え時
        $slickContainer.on('beforeChange', function(event, slick, currentSlide, nextSlide) {
            if (currentSlide !== nextSlide) {
                if ($vdiffjs.is(':visible')) {
                    showDetail(nextSlide);
                }
            }
        });

        function showDetail(index) {
            var enableEdit = config__.enableEdit;
            if (enableEdit) {
                showDetailWithEditor(index);
            } else {
                $vdiffjs.empty();
                if (typeof VDiff !== 'undefined') {
                    VDiff({
                        id: vdiffjsId,
                        data: resultSequence[index],
                        lang: lng
                    });
                }
            }
        }
        function showDetailWithEditor(index) {
            $vdiffjs.empty();
            var editor;
            var isEditorMode = false;
            var _viewerId = '#' + vdiffjsId;
            var _editorId = '#' + vdiffjsEditorId;
            var data_ = resultSequence[index];
            var viewerConfig = {
                id: vdiffjsId,
                data: data_,
                editorUrl: 'javascript:void(0);',
                lang: lng
            };
            var editorConfig = {
                id: vdiffjsEditorId,
                layout: 'horizontal',
                showSaveParametersButton: true,
                lang: lng,
                doneCallback: function() {
                    $(_editorId + '-save_param').attr({ title: (lng !== 'ja') ? 'Back to image comparison' : '画像比較に戻る' });
                }
            };
            function updateVE() {
                if (isEditorMode) {
                    //編集モード
                    $(_editorId).removeAttr('class').removeAttr('style').empty().show();
                    $(_viewerId).removeAttr('class').removeAttr('style').empty().hide();
                    editorConfig.data = data_;
                    if (typeof VDiffEditor !== 'undefined') {
                        editor = VDiffEditor(editorConfig);
                    }
                } else {
                    //比較表示モード
                    if (editor) {
                        //editorがあれば、editorから編集後のパラメータを取得
                        data_ = editor.getParameters() || data_;
                        console.log(JSON.stringify(resultSequence)); // eslint-disable-line no-console
                    }
                    $(_viewerId).removeAttr('class').removeAttr('style').empty().show();
                    $(_editorId).removeAttr('class').removeAttr('style').empty().hide();
                    viewerConfig.data = data_;
                    if (typeof VDiff !== 'undefined') {
                        VDiff(viewerConfig);
                    }
                }
                isEditorMode = !(isEditorMode);
            }
            $(_viewerId).on('click', _viewerId + '-editor_link', updateVE);
            $(_editorId).on('click', _editorId + '-save_param', updateVE);

            updateVE();
        }

        //累積的射影変換のリセットボタン
        var resetAccumulationButtonLabel = '<i class="fas fa-thumbtack"></i>';
        var $resetAccumulationButton = $('<button>').html(resetAccumulationButtonLabel).addClass('ui-button ui-widget ui-corner-all').attr({ type: 'button', title: (lng !== 'ja') ? 'Realigning comparison results' : '比較結果の合わせ直し' });
        $detailButtonWrapper.append($resetAccumulationButton);
        $resetAccumulationButton.on('click', function() {
            var currentSlide = $slickContainer.slick('slickCurrentSlide');
            resetAccumulatedImagesAfter(currentSlide);
        });
        function resetAccumulatedImagesAfter(index) {
            //index以降の射影変換画像を更新する
            for (var i = index; i < $imgs.length - 1; i++) {
                var cumulativeHomographyMatrix = (i > 0) ? matFromArrayEx(cumulativeHomographyArrays[i - 1]) : new cv.Mat();
                var transMatrix = matFromArrayEx(transArrays[i]);

                //cv.imreadは画像の表示サイズに左右される（ただし、読み込み画像が非表示状態のときは、naturalWidth, naturalHeightサイズとなる）
                var img = cv.imread($canvasOutputNoncumulativeArray[i][0].id);
                var work = new cv.Mat();
                if (i === index) {
                    var thisImageSize = new cv.Size(img.cols, img.rows);
                    var ratio = Math.max(thisImageSize.width / firstImageSize.width, thisImageSize.height / firstImageSize.height);
                    if (ratio <= 0) { ratio = 1; }

                    //firstImageSizeサイズに内接するように拡大縮小する
                    var quad1 = cv.matFromArray(4, 1, cv.CV_32FC2, [0, 0, 0, ratio, ratio, 0, ratio, ratio]);
                    var quad2 = cv.matFromArray(4, 1, cv.CV_32FC2, [0, 0, 0, 1, 1, 0, 1, 1]);
                    var toFirstImgSizeTransMat = cv.getPerspectiveTransform(quad1, quad2);
                    cv.warpPerspective(img, work, toFirstImgSizeTransMat, firstImageSize, cv.INTER_LINEAR, cv.BORDER_CONSTANT);
                    quad1.delete(); quad2.delete();

                    cv.imshow($canvasOutputArray[i][0].id, work);

                    //cumulativeHomographyMatrixを更新
                    cv.gemm(toFirstImgSizeTransMat, transMatrix, 1, new cv.Mat(), 0, cumulativeHomographyMatrix);
                    toFirstImgSizeTransMat.delete();
                } else {
                    cv.warpPerspective(img, work, cumulativeHomographyMatrix, firstImageSize, cv.INTER_LINEAR, cv.BORDER_CONSTANT);
                    cv.imshow($canvasOutputArray[i][0].id, work);

                    //cumulativeHomographyMatrixを更新
                    cv.gemm(cumulativeHomographyMatrix, transMatrix, 1, new cv.Mat(), 0, cumulativeHomographyMatrix);
                }
                work.delete();
                img.delete();

                //cumulativeHomographyMatrixを更新
                cumulativeHomographyArrays[i] = arrayExFromMat(cumulativeHomographyMatrix);

                cumulativeHomographyMatrix.delete();
                transMatrix.delete();
            }
        }

        //外部リンクボタン
        var extlinkButtonLabel = '<i class="fas fa-external-link-alt"></i>';
        var $extlinkButton = $('<a>').html(extlinkButtonLabel).addClass('ui-button ui-widget ui-corner-all').attr({ title: (lng !== 'ja') ? 'Related information' : '関連情報' });
        $detailButtonWrapper.append($extlinkButton);
        function updateExtlinkButton(index) {
            var data = resultSequence[index];
            var url;
            if (data && data.related && $.type(data.related) === 'string') {
                url = data.related;
            }
            if (url) {
                $extlinkButton.attr({ href: url, target: '_blank' }).show();
            } else {
                $extlinkButton.hide();
            }
        }
        $slickContainer.on('afterChange', function(event, slick, currentSlide) {
            updateExtlinkButton(currentSlide);
        });
        $slickContainer.on('init', function() {
            updateExtlinkButton(0);
        });

        //ヘルプボタン
        var helpButtonLabel = '<i class="fas fa-question"></i>';
        var $helpButton = $('<a>').html(helpButtonLabel).addClass('ui-button ui-widget ui-corner-all').attr({ href: 'http://codh.rois.ac.jp/software/vdiffseqjs/', target: '_blank', title: (lng !== 'ja') ? 'Help about vdiff-seq.js' : 'vdiff-seq.jsヘルプ' });
        $detailButtonWrapper.append($helpButton);

        // マルチスレッド処理ではないので問題ない
        var cumulativeHomographyMatrix = new cv.Mat();
        var resultSequence = [];

        var cumulativeHomographyArrays = [];
        var transArrays = [];

        var firstImageSize; //リサイズ（読み込みサイズ）座標系（整数）
        var firstImageDispWidth; //表示座標系

        function arrayExFromMat(mat) { //1ch matrixのみ対応
            if (mat.channels() !== 1) {
                return null;
            }
            var depth = mat.depth();
            var array = [];
            for (var i = 0; i < mat.rows; i++) {
                for (var j = 0; j < mat.cols; j++) {
                    switch (depth) {
                    case cv.CV_8U:
                        array.push(mat.ucharPtr(i, j)[0]);
                        break;
                    case cv.CV_8S:
                        array.push(mat.charPtr(i, j)[0]);
                        break;
                    case cv.CV_16U:
                        array.push(mat.ushortPtr(i, j)[0]);
                        break;
                    case cv.CV_16S:
                        array.push(mat.shortPtr(i, j)[0]);
                        break;
                    case cv.CV_32S:
                        array.push(mat.intPtr(i, j)[0]);
                        break;
                    case cv.CV_32F:
                        array.push(mat.floatPtr(i, j)[0]);
                        break;
                    case cv.CV_64F:
                        array.push(mat.doublePtr(i, j)[0]);
                        break;
                    }
                }
            }
            // console.log(array);
            return {
                rows: mat.rows,
                cols: mat.cols,
                type: mat.type(),
                array: array
            };
        }
        function matFromArrayEx(arrayEx) { //1ch matrixのみ対応
            if (arrayEx && arrayEx.rows && arrayEx.cols && 'type' in arrayEx && $.isArray(arrayEx.array)) {
                if (arrayEx.rows * arrayEx.cols === arrayEx.array.length) {
                    return cv.matFromArray(arrayEx.rows, arrayEx.cols, arrayEx.type, arrayEx.array);
                }
            }
            return null;
        }

        function showComparedImagesCore() {
            function processOneData(i) {
                var d = $.Deferred();
                setTimeout(function() {
                    try {
                        var data_ = JSON.parse(JSON.stringify(sequence[i]));
                        var index1 = imgUrlDict.indexOf(data_.img1);
                        var index2 = imgUrlDict.indexOf(data_.img2);
                        if (index1 < 0 || index2 < 0) {
                            d.reject();
                        }
                        var returnValues = showComparedImagesCoreCore(data_, $imgs[index1], $imgs[index2], $canvasOutputArray[i], $canvasOutputNoncumulativeArray[i], i);
                        data_['corr_pts'] = JSON.parse(JSON.stringify(returnValues.corrPts)); //更新
                        if (isGlobalCompareMode) {
                            //全体変化比較モード
                            data_.mode = true;
                        }
                        var $slickSlide = $('<li>');
                        $canvasOutputArray[i].appendTo($slickSlide);

                        if (data_.img1_label && data_.img2_label) {
                            if (returnValues && 'isBlackLetterOnWhitePaper' in returnValues) {
                                var $label = $('<div>').addClass('slick-label');

                                var $legendMarkerL = $('<div>').addClass('legend-marker');
                                var $legendMarkerR = $('<div>').addClass('legend-marker');

                                var $label1 = $('<span>').text(data_.img1_label);
                                var $label2 = $('<span>').text(data_.img2_label);
                                if (returnValues.isBlackLetterOnWhitePaper === true) {
                                    //どちらの画像も白地に黒文字であるとき、
                                    //画像1にはあって画像2にはない文字部分が青 → 青は画像1（左側画像）
                                    //画像2にはあって画像1にはない文字部分が赤 → 赤は画像2（右側画像）
                                    $legendMarkerL.addClass('legend-blue');
                                    $legendMarkerR.addClass('legend-red');
                                } else if (returnValues.isBlackLetterOnWhitePaper === false) {
                                    //どちらの画像も黒地に白文字であるとき、
                                    //画像1にはあって画像2にはない文字部分が赤 → 赤は画像1（左側画像）
                                    //画像2にはあって画像1にはない文字部分が青 → 青は画像2（右側画像）
                                    $legendMarkerL.addClass('legend-red');
                                    $legendMarkerR.addClass('legend-blue');
                                } else {
                                    $legendMarkerL.addClass('legend-gray');
                                    $legendMarkerR.addClass('legend-gray');
                                }
                                $label.append($legendMarkerL).append($label1).append('<br>').append($legendMarkerR).append($label2);
                                $slickSlide.append($label);
                            }
                        }
                        $slickContainer.append($slickSlide);
                        $slickSlide.hide();

                        resultSequence.push(data_);

                        if (sequence.length > 0) {
                            $progressbar.progressbar('value', Math.round((i + 1) / sequence.length * 100));
                        }

                        d.resolve();
                    } catch(e) {
                        d.reject();
                    }
                }, 0);
                return d;
            }
            function getExportData() {
                var currentSlide = $slickContainer.slick('slickCurrentSlide');
                var data = resultSequence[currentSlide];
                return data;
            }
            function getSequence() {
                return resultSequence;
            }
            $imagesContainer.show();
            if ($imgs.length > 0) {
                var $firstImageDiv = $imagesContainer.children('div').first();
                if ($firstImageDiv) {
                    $(containerId).css({
                        'width': $firstImageDiv.width(),
                        'height': $firstImageDiv.height(),
                        'box-sizing': 'content-box',
                        'overflow': 'hidden'
                    });
                }
            }
            var deferreds = [];
            for (var i = 0; i < sequence.length; i++) {
                deferreds.push(processOneData(i));
            }
            function showComparedImagesCoreDone() {
                cumulativeHomographyMatrix.delete();

                $imagesContainer.remove();
                $progressbarContainer.remove();
                $outputContainer.remove();
                if (firstImageDispWidth) {
                    $(containerId).css({ width: firstImageDispWidth });
                }
                $(containerId).css({ height: '', overflow: '' });

                $slickContainer.find('li').show();
                $slickContainer.slick({
                    dots: true,
                    speed: 0,
                    infinite: false,
                    draggable: false,
                    swipe: false
                });
                $slickContainer.after($detailButtonWrapper);

                if (params__.debug) {
                    console.log(JSON.stringify(getSequence())); // eslint-disable-line no-console
                }

                //API提供
                exportFunctions = {
                    getParameters: getExportData,
                    getSequence: getSequence
                };

                //正常完了
                if ($.isFunction(config.doneCallback)) {
                    config.doneCallback();
                }
                if ($.isFunction(config.alwaysCallback)) {
                    config.alwaysCallback();
                }
            }
            if (sequence.length > 0) {
                $progressbarContainer.css({ width: $imgs[0][0].width, height: $imgs[0][0].height, top: 0 });
            }
            $progressbar.progressbar({
                value: false,
                change: function() {
                    $progressLabel.text($progressbar.progressbar('value') + '%');
                },
                complete: function() {
                    $progressLabel.text('');
                    $progressbar.hide();
                }
            });
            $.when.apply($, deferreds)
                .done(showComparedImagesCoreDone)
                .fail(function() {
                    showError((lng !== 'ja') ? 'Image comparison could not be completed.' : '画像比較でエラーが発生しました。');
                });
        }
        function showComparedImagesCoreCore(data_, $img1, $img2, $canvasOutput, $canvasOutputNoncumulative, index) {
            var isThisBaseFrame = (index === 0);

            var returnValues = {};
            var instances = [];

            //画像サイズの関係：
            //フルサイズ（オリジナル）座標系 →（'max-size'設定）→ リサイズ（読み込みサイズ）座標系 →（'max-width'/'max-height'設定）→ 表示座標系

            //画像サイズ制限（'max-size'設定）
            var MAX_SIZE_DEFAULT = 2000; //px
            var maxSize = (limitSize && $.isPlainObject(limitSize) && $.isNumeric(limitSize['max-size'])) ? limitSize['max-size'] : MAX_SIZE_DEFAULT;
            if (maxSize <= 0) {
                maxSize = MAX_SIZE_DEFAULT;
            }
            function getSizeLimitInfo(img, maxSize) {
                var ratio = 1;
                if (img.naturalWidth > img.naturalHeight) {
                    if (img.naturalWidth > maxSize) {
                        ratio = maxSize / img.naturalWidth;
                        return { css: { width: maxSize }, ratio: ratio, resized: { width: maxSize, height: img.naturalHeight * ratio } };
                    }
                } else {
                    if (img.naturalHeight > maxSize) {
                        ratio = maxSize / img.naturalHeight;
                        return { css: { height: maxSize }, ratio: ratio, resized: { width: img.naturalWidth * ratio, height: maxSize } };
                    }
                }
                return { css: {}, ratio: ratio, resized: { width: img.naturalWidth * ratio, height: img.naturalHeight * ratio} };
            }
            var img1SizeLimit = getSizeLimitInfo($img1[0], maxSize);
            var img2SizeLimit = getSizeLimitInfo($img2[0], maxSize);
            $img1.css(img1SizeLimit.css);
            $img2.css(img2SizeLimit.css);

            //'max-size'設定を反映したサイズで画像を読み込むため、いったん移動する
            var $img1Wrapper = $('<div>').css({ width: $img1[0].naturalWidth, height: $img1[0].naturalHeight, padding: 0 });
            var $img2Wrapper = $('<div>').css({ width: $img2[0].naturalWidth, height: $img2[0].naturalHeight, padding: 0 });
            $(containerId).append($img1Wrapper).append($img2Wrapper);
            $img1.appendTo($img1Wrapper);
            $img2.appendTo($img2Wrapper);
            //cv.imreadは画像の表示サイズに左右される（ただし、読み込み画像が非表示状態のときは、naturalWidth, naturalHeightサイズとなる）
            var img1 = cv.imread($img1[0]); instances.push(img1); //リサイズ座標系
            var img2 = cv.imread($img2[0]); instances.push(img2); //リサイズ座標系
            //フルサイズ座標系
            var img1SizeNatural = {
                width:  $img1[0].width,
                height: $img1[0].height
            };
            var img2SizeNatural = {
                width:  $img2[0].width,
                height: $img2[0].height
            };
            $img1Wrapper.remove();
            $img2Wrapper.remove();

            var dispW = $img1[0].width;  //フルサイズ座標系 → ... → 表示座標系（'max-size', 'max-width'または'max-height'設定の影響を受けうる）
            var dispH = $img1[0].height; //フルサイズ座標系 → ... → 表示座標系（'max-size', 'max-width'または'max-height'設定の影響を受けうる）
            var dispRatio = 1; //cv.imreadで読み込んだサイズから表示サイズへの倍率（縮小率）
            var dispSizeCss;
            //フルサイズ座標系 →（'max-size'設定）→ リサイズ座標系
            if (dispW > maxSize || dispH > maxSize) {
                //dispRatioの設定は不要
                dispW = img1SizeLimit.resized.width;
                dispH = img1SizeLimit.resized.height;
                dispSizeCss = { width: dispW, height: dispH, 'max-width': dispW, 'max-height': dispH };
            }
            //リサイズ座標系 →（'max-width'/'max-height'設定）→ 表示座標系
            if (limitSize && $.isPlainObject(limitSize)) {
                var dispRatio_;
                if ($.isNumeric(limitSize['max-width']) && limitSize['max-width'] > 0 && dispW > 0) {
                    dispRatio_ = limitSize['max-width'] / dispW;
                    if (dispRatio_ < 1) {
                        dispRatio = dispRatio_;
                        dispW = limitSize['max-width'];
                        dispH *= dispRatio;
                        dispSizeCss = { width: dispW, height: dispH, 'max-width': dispW, 'max-height': dispH };
                    }
                } else if ($.isNumeric(limitSize['max-height']) && limitSize['max-height'] > 0 && dispH > 0) {
                    dispRatio_ = limitSize['max-height'] / dispH;
                    if (dispRatio_ < 1) {
                        dispRatio = dispRatio_;
                        dispH = limitSize['max-height'];
                        dispW *= dispRatio;
                        dispSizeCss = { width: dispW, height: dispH, 'max-width': dispW, 'max-height': dispH };
                    }
                }
            }

            function getRectFromFragment(fragment) {
                var rect = { x: 0, y: 0, width: 100, height: 100, xywhUnit: 'percent' };
                if (fragment) {
                    //https://www.w3.org/TR/media-frags/#naming-space
                    //Media Fragmentsでは「pixel:」「percent:」（省略時はpixel扱い）、値は整数
                    var match = fragment.match(/^(pixel:|percent:)?([0-9]+),([0-9]+),([0-9]+),([0-9]+)/);
                    if (!match) {
                        //https://iiif.io/api/image/2.1/#region
                        //IIIF Image APIのregion指定では「pct:」または省略、「pct:」指定時の値は正の浮動小数点数または整数
                        //https://www.wikidata.org/wiki/Property:P2677
                        //pct:((100|[1-9]?\d(\.\d+)?),){3}(100|[1-9]?\d(\.\d+)?)
                        match = fragment.match(/^(pct:)(100|[1-9]?\d(?:\.\d+)?),(100|[1-9]?\d(?:\.\d+)?),(100|[1-9]?\d(?:\.\d+)?),(100|[1-9]?\d(?:\.\d+)?)/);
                    }
                    if (match) {
                        var x, y, w, h, xywhUnit;
                        if (match[1] === 'percent:' || match[1] === 'pct:') {
                            xywhUnit = 'percent';
                        }
                        if (xywhUnit === 'percent') {
                            x = parseFloat(match[2]);
                            y = parseFloat(match[3]);
                            w = parseFloat(match[4]);
                            h = parseFloat(match[5]);
                        } else {
                            x = parseInt(match[2], 10);
                            y = parseInt(match[3], 10);
                            w = parseInt(match[4], 10);
                            h = parseInt(match[5], 10);
                        }
                        rect = { x: x, y: y, width: w, height: h, xywhUnit: xywhUnit };
                    }
                }
                return rect;
            }
            function getIntersectRect(width, height, rect, unit) {
                var intersectRect;
                var x = rect.x;
                var y = rect.y;
                var w = rect.width;
                var h = rect.height;
                var xmin, xmax;
                var ymin, ymax;
                var swap;
                if (unit === 'percent') {
                    xmin = Math.min(Math.max(0, x), 100);
                    xmax = Math.max(Math.min(100, x + w), 0);
                    ymin = Math.min(Math.max(0, y), 100);
                    ymax = Math.max(Math.min(100, y + h), 0);
                } else {
                    xmin = Math.min(Math.max(0, x), width);
                    xmax = Math.max(Math.min(width, x + w), 0);
                    ymin = Math.min(Math.max(0, y), height);
                    ymax = Math.max(Math.min(height, y + h), 0);
                }
                if (xmin > xmax) {
                    swap = xmin;
                    xmin = xmax;
                    xmax = swap;
                }
                if (ymin > ymax) {
                    swap = ymin;
                    ymin = ymax;
                    ymax = swap;
                }
                if (unit === 'percent') {
                    intersectRect = {
                        x: xmin / 100 * width,
                        y: ymin / 100 * height,
                        width: (xmax - xmin) / 100 * width,
                        height: (ymax - ymin)  / 100 * height
                    };
                } else {
                    intersectRect = {
                        x: xmin,
                        y: ymin,
                        width: (xmax - xmin),
                        height: (ymax - ymin)
                    };
                }
                return intersectRect;
            }
            var dsize = new cv.Size(img1.cols, img1.rows); //出力画像はimg1と同サイズ（リサイズ座標系）
            var drect = new cv.Rect(0, 0, img1.cols, img1.rows); //ROI矩形（リサイズ座標系）
            if (isThisBaseFrame) {
                firstImageSize = dsize; //リサイズ（読み込みサイズ）座標系
                firstImageDispWidth = dispW; //表示座標系
            }
            if (data_.img1_roi_xywh) {
                var rectRoi = getRectFromFragment(data_.img1_roi_xywh); //フルサイズ座標系または相対値
                if (img1SizeLimit.ratio < 1 && rectRoi.xywhUnit !== 'percent') {
                    //'max-size'設定により、img1が縮小されている場合、img1_roi_xywhに縮小率を反映させる
                    //フルサイズ座標系 → リサイズ座標系
                    rectRoi.x *= img1SizeLimit.ratio;
                    rectRoi.y *= img1SizeLimit.ratio;
                    rectRoi.width *= img1SizeLimit.ratio;
                    rectRoi.height *= img1SizeLimit.ratio;
                }
                drect = getIntersectRect(img1.cols, img1.rows, rectRoi, rectRoi.xywhUnit);
                if (drect.x === 0 && drect.y === 0 && drect.width === 0 & drect.height === 0) {
                    drect.width = img1.cols;
                    drect.height = img1.rows;
                }
            }

            //---
            //img2を射影変換
            //---
            var img2Warped = new cv.Mat(); instances.push(img2Warped);
            var img2WarpedMask = new cv.Mat(); instances.push(img2WarpedMask);
            var transMatrix;
            (function() {
                function validateCorrespondingPointsInput(data_) {
                    try {
                        var data;
                        if ($.isPlainObject(data_)) {
                            data = data_;
                        } else {
                            data = JSON.parse(data_);
                        }
                        if ($.isPlainObject(data) &&
                            $.isPlainObject(data.img1) && $.isPlainObject(data.img2) &&
                            $.isPlainObject(data.img1.size) && $.isPlainObject(data.img2.size) &&
                            Number.isFinite(data.img1.size.w) && Number.isFinite(data.img1.size.h) &&
                            Number.isFinite(data.img2.size.w) && Number.isFinite(data.img2.size.h) &&
                            data.img1.size.w > 0 && data.img1.size.h > 0 &&
                            data.img2.size.w > 0 && data.img2.size.h > 0 &&
                            $.isArray(data.img1.pts) && $.isArray(data.img2.pts) &&
                            data.img1.pts.length === data.img2.pts.length &&
                            data.img1.pts.length > 3) {
                            for (var i; i < data.img1.pts.length; i++) {
                                var pt1 = data.img1.pts[i];
                                var pt2 = data.img2.pts[i];
                                if ($.isPlainObject(pt1) && $.isPlainObject(pt2) &&
                                    Number.isFinite(pt1.x) && Number.isFinite(pt1.y) &&
                                    Number.isFinite(pt2.x) && Number.isFinite(pt2.y)) {
                                    //OK
                                } else {
                                    return false;
                                }
                            }
                            return data;
                        }
                    } catch(e) {
                        //有効なJSONでない
                    }
                    return false;
                }
                function getCornerPts(imgMat) {
                    var pts = [];
                    pts.push({ x: 0, y: 0 });
                    pts.push({ x: imgMat.cols, y: 0 });
                    pts.push({ x: imgMat.cols, y: imgMat.rows });
                    pts.push({ x: 0, y: imgMat.rows });
                    return pts;
                }
                var corrPts_ = validateCorrespondingPointsInput(data_.corr_pts);
                if (corrPts_ || data_.corr_pts === 'corners') {
                    transMatrix = (function() {
                        var pts1 = []; //リサイズ座標系
                        var pts2 = []; //リサイズ座標系
                        if (corrPts_) {
                            //与えられた対応点情報を用いる
                            var size1 = corrPts_.img1.size;
                            var size2 = corrPts_.img2.size;
                            var pts1_ = corrPts_.img1.pts;
                            var pts2_ = corrPts_.img2.pts;
                            for (var i = 0; i < pts1_.length; i++) {
                                //リサイズ座標系
                                pts1.push({
                                    x: pts1_[i].x / size1.w * img1.cols,
                                    y: pts1_[i].y / size1.h * img1.rows,
                                });
                                pts2.push({
                                    x: pts2_[i].x / size2.w * img2.cols,
                                    y: pts2_[i].y / size2.h * img2.rows,
                                });
                            }
                        } else if (data_.corr_pts === 'corners') {
                            //各画像の四隅を対応点として用いる
                            //リサイズ座標系
                            pts1 = getCornerPts(img1);
                            pts2 = getCornerPts(img2);
                        }
                        var quad1 = cv.matFromArray(4, 1, cv.CV_32FC2,
                            [pts1[0].x, pts1[0].y, pts1[1].x, pts1[1].y, pts1[2].x, pts1[2].y, pts1[3].x, pts1[3].y]);
                        var quad2 = cv.matFromArray(4, 1, cv.CV_32FC2,
                            [pts2[0].x, pts2[0].y, pts2[1].x, pts2[1].y, pts2[2].x, pts2[2].y, pts2[3].x, pts2[3].y]);
                        if (!data_.img1_roi_xywh) {
                            //ROIが明示的に指定されていなければ、画像1の対応点列の外接矩形をROIとする
                            var boundingRect = cv.boundingRect(quad1); //リサイズ座標系
                            drect = getIntersectRect(img1.cols, img1.rows, boundingRect); //リサイズ座標系
                            if (drect.x === 0 && drect.y === 0 && drect.width === 0 & drect.height === 0) {
                                drect.width = img1.cols;
                                drect.height = img1.rows;
                            }
                        }
                        var perspectiveMatrix = cv.getPerspectiveTransform(quad2, quad1);
                        quad1.delete(); quad2.delete();
                        return perspectiveMatrix;
                    }()); instances.push(transMatrix);
                } else {
                    transMatrix = (function() {
                        //対応点を自動推定する
                        var instances_ = [];
                        var mask1 = new cv.Mat(); instances_.push(mask1);
                        var mask2 = new cv.Mat(); instances_.push(mask2);

                        var roi = new cv.Mat(drect.height, drect.width, cv.CV_8UC1, new cv.Scalar(255)); instances_.push(roi);
                        var transMat = cv.matFromArray(2, 3, cv.CV_64FC1, [1, 0, drect.x, 0, 1, drect.y]); instances_.push(transMat);
                        cv.warpAffine(roi, mask1, transMat, dsize, cv.INTER_LINEAR, cv.BORDER_CONSTANT); //リサイズ座標系

                        var akaze = new cv.AKAZE(); instances_.push(akaze);
                        var kp1 = new cv.KeyPointVector(); instances_.push(kp1);
                        var kp2 = new cv.KeyPointVector(); instances_.push(kp2);
                        var descriptors1 = new cv.Mat(); instances_.push(descriptors1);
                        var descriptors2 = new cv.Mat(); instances_.push(descriptors2);
                        akaze.detectAndCompute(img1, mask1, kp1, descriptors1, false); //リサイズ座標系
                        akaze.detectAndCompute(img2, mask2, kp2, descriptors2, false); //リサイズ座標系

                        var matches = new cv.DMatchVectorVector(); instances_.push(matches);
                        var matcher = new cv.BFMatcher(cv.NORM_HAMMING); instances_.push(matcher);
                        matcher.knnMatch(descriptors1, descriptors2, matches, 2);

                        var RATIO_TEST_THRESHOLD = 0.5;
                        var matchPoint1 = [];
                        var matchPoint2 = [];
                        for (var i = 0; i < matches.size(); i++) {
                            var m = matches.get(i).get(0);
                            var n = matches.get(i).get(1);
                            if (m.distance < RATIO_TEST_THRESHOLD * n.distance) {
                                var kpt1 = kp1.get(m.queryIdx).pt;
                                var kpt2 = kp2.get(m.trainIdx).pt;
                                //リサイズ座標系
                                matchPoint1.push(kpt1.x);
                                matchPoint1.push(kpt1.y);
                                matchPoint2.push(kpt2.x);
                                matchPoint2.push(kpt2.y);
                            }
                        }

                        var srcPoints = cv.matFromArray(matchPoint1.length / 2, 1, cv.CV_32FC2, matchPoint1); instances_.push(srcPoints);
                        var dstPoints = cv.matFromArray(matchPoint2.length / 2, 1, cv.CV_32FC2, matchPoint2); instances_.push(dstPoints);
                        var homographyMatrix; //リサイズ座標系
                        if (matchPoint1.length / 2 >= 4 && matchPoint1.length / 2 >= 4) {
                            homographyMatrix = cv.findHomography(dstPoints, srcPoints, cv.RANSAC, 3.0);
                        } else {
                            //各画像の四隅を対応点として用いる
                            //リサイズ座標系
                            var pts1 = getCornerPts(img1);
                            var pts2 = getCornerPts(img2);
                            var quad1 = cv.matFromArray(4, 1, cv.CV_32FC2,
                                [pts1[0].x, pts1[0].y, pts1[1].x, pts1[1].y, pts1[2].x, pts1[2].y, pts1[3].x, pts1[3].y]);
                            var quad2 = cv.matFromArray(4, 1, cv.CV_32FC2,
                                [pts2[0].x, pts2[0].y, pts2[1].x, pts2[1].y, pts2[2].x, pts2[2].y, pts2[3].x, pts2[3].y]);
                            homographyMatrix = cv.getPerspectiveTransform(quad2, quad1);
                            quad1.delete(); quad2.delete();
                        }
                        instances_.forEach(function(m) { m.delete(); });
                        return homographyMatrix;
                    }()); instances.push(transMatrix);
                }
                var img2Flat = new cv.Mat(img2.rows, img2.cols, cv.CV_8UC1, new cv.Scalar(255)); //img2と同じ大きさ
                cv.warpPerspective(img2, img2Warped, transMatrix, dsize, cv.INTER_LINEAR, cv.BORDER_CONSTANT); //img1の大きさに合わせる
                cv.warpPerspective(img2Flat, img2WarpedMask, transMatrix, dsize, cv.INTER_LINEAR, cv.BORDER_CONSTANT); //img1の大きさに合わせる
                img2Flat.delete();
            }());

            //unsharp処理
            var img2WarpedUnsharped = new cv.Mat(); instances.push(img2WarpedUnsharped);
            var work1 = new cv.Mat(); instances.push(work1);
            cv.GaussianBlur(img2Warped, work1, new cv.Size(3, 3), 1);
            cv.addWeighted(img2Warped, 1.5, work1, -0.5, 0, img2WarpedUnsharped);

            if (isGlobalCompareMode) {
                //全体変化比較モード

                //半透明重ね合わせ
                var imgMerged_ = new cv.Mat(); instances.push(imgMerged_);
                cv.addWeighted(img1, 0.5, img2WarpedUnsharped, 0.5, 0, imgMerged_); //同じサイズでなければならない

                if (isThisBaseFrame) {
                    cv.imshow($canvasOutput[0].id, imgMerged_);
                    cv.imshow($canvasOutputNoncumulative[0].id, imgMerged_);
                } else {
                    cv.warpPerspective(imgMerged_, work1, cumulativeHomographyMatrix, firstImageSize, cv.INTER_LINEAR, cv.BORDER_CONSTANT);
                    cv.imshow($canvasOutput[0].id, work1);
                    cv.imshow($canvasOutputNoncumulative[0].id, imgMerged_);
                }
                returnValues.isBlackLetterOnWhitePaper = undefined;

            } else {
                //局所強調比較モード

                //----------
                //ハイライト画像の作成
                //---

                //二値化した入力画像を用いてノイズを削減するオプション
                var nrbi = true; //noise reduction using binarized input images
                if ('nrbi' in params__) {
                    nrbi = (params__.nrbi === '1' || params__.nrbi.toLowerCase() === 'true');
                }

                //半透明重ね合わせ
                var imgMerged = new cv.Mat(); instances.push(imgMerged);
                cv.addWeighted(img1, 0.5, img2WarpedUnsharped, 0.5, 0, work1); //同じサイズでなければならない
                cv.cvtColor(work1, imgMerged, cv.COLOR_RGB2RGBA, 0); //RGBA

                var work2 = new cv.Mat(); instances.push(work2);
                var work3 = new cv.Mat(); instances.push(work3);
                var work4 = new cv.Mat(); instances.push(work4);
                var work5 = new cv.Mat(); instances.push(work5);
                var work6 = new cv.Mat(); instances.push(work6);
                var work7 = new cv.Mat(); instances.push(work7);

                //グレイスケールに変換
                cv.cvtColor(img1,       work1, cv.COLOR_RGBA2GRAY, 0);
                cv.cvtColor(img2Warped, work2, cv.COLOR_RGBA2GRAY, 0);

                if (nrbi) {
                    //二値化（入力画像は白地に黒文字を想定）
                    cv.threshold(work1, work3, 0, 255, cv.THRESH_OTSU);
                    cv.threshold(work2, work4, 0, 255, cv.THRESH_OTSU);
                    //マスクを作成（二値化入力画像の共通部分は黒、相違部分は白になる）
                    cv.bitwise_xor(work3, work4, work7);

                    if (work3.rows * work3.cols > 0 && work4.rows * work4.cols > 0) {
                        //二値化入力画像に基づき、入力画像と強調表示色の対応関係を推測
                        var WHITE_RATE_THRESHOLD = 0.5;
                        var whiteRate1 = cv.countNonZero(work3) / (work3.rows * work3.cols); //img1
                        var whiteRate2 = cv.countNonZero(work4) / (work4.rows * work4.cols); //img2
                        if (whiteRate1 >= WHITE_RATE_THRESHOLD && whiteRate2 >= WHITE_RATE_THRESHOLD) {
                            //どちらの画像も白地に黒文字であるとき
                            returnValues.isBlackLetterOnWhitePaper = true;
                        } else if (whiteRate1 < WHITE_RATE_THRESHOLD && whiteRate2 < WHITE_RATE_THRESHOLD) {
                            //どちらの画像も黒地に白文字であるとき
                            returnValues.isBlackLetterOnWhitePaper = false;
                        } else {
                            returnValues.isBlackLetterOnWhitePaper = undefined;
                        }
                    }
                }

                //差分（元画像の黒文字が白、白地が黒になった状態）
                cv.subtract(work1, work2, work3);
                cv.subtract(work2, work1, work4);
                cv.absdiff(work1, work2, work5);

                //ノイズ除去（メディアンフィルタ）
                cv.medianBlur(work3, work1, 3);
                cv.medianBlur(work4, work2, 3);
                cv.medianBlur(work5, work6, 3);

                //二値化
                cv.threshold(work1, work3, 0, 255, cv.THRESH_OTSU);
                cv.threshold(work2, work4, 0, 255, cv.THRESH_OTSU);
                cv.threshold(work6, work5, 0, 255, cv.THRESH_OTSU);

                if (nrbi) {
                    //二値化された入力画像で共通している部分はマスクする
                    cv.bitwise_and(work3, work7, work3);
                    cv.bitwise_and(work4, work7, work4);
                    cv.bitwise_and(work5, work7, work5);
                }

                //ROI矩形外部は黒で塗りつぶし
                var roiWork3 = work3.roi(drect); instances.push(roiWork3);
                var roiWork4 = work4.roi(drect); instances.push(roiWork4);
                var roiWork5 = work5.roi(drect); instances.push(roiWork5);
                var transMat = cv.matFromArray(2, 3, cv.CV_64FC1, [1, 0, drect.x, 0, 1, drect.y]); instances.push(transMat);
                cv.warpAffine(roiWork3, work1, transMat, dsize, cv.INTER_LINEAR, cv.BORDER_CONSTANT);
                cv.warpAffine(roiWork4, work2, transMat, dsize, cv.INTER_LINEAR, cv.BORDER_CONSTANT);
                cv.warpAffine(roiWork5, work6, transMat, dsize, cv.INTER_LINEAR, cv.BORDER_CONSTANT);

                //img2を射影変換した外側の領域についてはマスクする
                cv.bitwise_and(work1, img2WarpedMask, work3);
                cv.bitwise_and(work2, img2WarpedMask, work4);
                cv.bitwise_and(work6, img2WarpedMask, work5);

                //ノイズ除去（メディアンフィルタ）
                cv.medianBlur(work3, work1, 3);
                cv.medianBlur(work4, work2, 3);
                cv.medianBlur(work5, work6, 3);

                //ノイズ除去（モルフォロジー変換）
                var ksize = 3;
                var iterations = 3;
                var kernel = cv.getStructuringElement(cv.MORPH_ELLIPSE, new cv.Size(ksize, ksize)); instances.push(kernel);
                cv.morphologyEx(work1, work3, cv.MORPH_OPEN , kernel); //OPEN: 縮めて拡大（白点ノイズ除去）
                cv.dilate(work3, work1, kernel, new cv.Point(-1, -1), iterations);
                cv.morphologyEx(work2, work4, cv.MORPH_OPEN , kernel); //OPEN: 縮めて拡大（白点ノイズ除去）
                cv.dilate(work4, work2, kernel, new cv.Point(-1, -1), iterations);
                cv.morphologyEx(work6, work5, cv.MORPH_OPEN , kernel); //OPEN: 縮めて拡大（白点ノイズ除去）
                cv.dilate(work5, work6, kernel, new cv.Point(-1, -1), iterations);

                //感度を下げる
                cv.bitwise_and(work6, work1, work3);
                cv.bitwise_and(work6, work2, work4);

                //ガウスぼかしで広げる
                cv.GaussianBlur(work3, work1, new cv.Size(3, 3), 1);
                cv.GaussianBlur(work4, work2, new cv.Size(3, 3), 1);

                //差分ハイライトマスク
                var rgbaPlanes1 = new cv.MatVector(); instances.push(rgbaPlanes1);
                var rgbaPlanes2 = new cv.MatVector(); instances.push(rgbaPlanes2);
                var plane0 = new cv.Mat(img1.rows, img1.cols, cv.CV_8UC1, new cv.Scalar(0)); instances.push(plane0);
                rgbaPlanes1.push_back(work1);
                rgbaPlanes1.push_back(plane0);
                rgbaPlanes1.push_back(plane0);
                rgbaPlanes1.push_back(work1);
                cv.merge(rgbaPlanes1, work3); //RGBA
                rgbaPlanes2.push_back(plane0);
                rgbaPlanes2.push_back(plane0);
                rgbaPlanes2.push_back(work2);
                rgbaPlanes2.push_back(work2);
                cv.merge(rgbaPlanes2, work4); //RGBA
                cv.addWeighted(work3, 1, work4, 1, 0, work5); //RGBA

                //差分ハイライトオーバーレイ画像
                cv.addWeighted(imgMerged, 1, work5, 1, 0, work6); //RGBA

                if (isThisBaseFrame) {
                    cv.imshow($canvasOutput[0].id, work6);
                    cv.imshow($canvasOutputNoncumulative[0].id, work6);
                } else {
                    cv.warpPerspective(work6, work5, cumulativeHomographyMatrix, firstImageSize, cv.INTER_LINEAR, cv.BORDER_CONSTANT);
                    cv.imshow($canvasOutput[0].id, work5);
                    cv.imshow($canvasOutputNoncumulative[0].id, work6);
                }
            }

            //corrPtsとcumulativeHomographyMatrixを更新
            var dispRatio1 = img1SizeLimit.ratio || 1;
            //リサイズ座標系 → フルサイズ座標系
            var drectNatural = {
                x: drect.x / dispRatio1,
                y: drect.y / dispRatio1,
                width: drect.width / dispRatio1,
                height: drect.height / dispRatio1
            };
            var drectNaturalPts = cv.matFromArray(4, 1, cv.CV_32FC2, [ //フルサイズ座標系
                drectNatural.x, drectNatural.y,
                drectNatural.x + drectNatural.width, drectNatural.y,
                drectNatural.x + drectNatural.width, drectNatural.y + drectNatural.height,
                drectNatural.x, drectNatural.y + drectNatural.height
            ]); instances.push(drectNaturalPts);
            var drectNaturalPtsTransformed = new cv.Mat(); instances.push(drectNaturalPtsTransformed);
            var transMatrixInv = new cv.Mat(); instances.push(transMatrixInv);
            cv.invert(transMatrix, transMatrixInv);
            cv.perspectiveTransform(drectNaturalPts, drectNaturalPtsTransformed, transMatrixInv);
            var img1Points = [];
            var img2Points = [];
            for (var j = 0; j < drectNaturalPtsTransformed.rows; j++) {
                var pt1 = drectNaturalPts.floatPtr(j, 0);
                img1Points.push({ x: pt1[0], y: pt1[1] }); //フルサイズ座標系
                var pt2 = drectNaturalPtsTransformed.floatPtr(j, 0);
                img2Points.push({ x: pt2[0], y: pt2[1] }); //フルサイズ座標系
            }
            //corrPtsを返り値に含める
            returnValues.corrPts = {
                //フルサイズ座標系
                img1: {
                    size: { w: img1SizeNatural.width, h: img1SizeNatural.height },
                    pts: img1Points
                },
                img2: {
                    size: { w: img2SizeNatural.width, h: img2SizeNatural.height },
                    pts: img2Points
                }
            };
            //cumulativeHomographyMatrixを更新
            if (isThisBaseFrame) {
                cumulativeHomographyMatrix = transMatrix.clone();
            } else {
                cv.gemm(cumulativeHomographyMatrix, transMatrix, 1, new cv.Mat(), 0, cumulativeHomographyMatrix);
            }
            cumulativeHomographyArrays.push(arrayExFromMat(cumulativeHomographyMatrix));
            transArrays.push(arrayExFromMat(transMatrix));

            if (dispSizeCss) {
                $(containerId + ' canvas').css(dispSizeCss);
            }

            //--- OpenCVの後始末
            instances.forEach(function(m) { m.delete(); });

            return returnValues;
        }

        //比較画像の取得後に処理
        function waitForLoading(index, img) {
            var d = $.Deferred();
            if (img.complete) {
                d.resolve();
            } else {
                $(img).on('load', function() { d.resolve(); });
                $(img).on('error', function() { d.reject(); });
            }
            return d;
        }
        $.when.apply($, $(containerId + ' img').map(waitForLoading))
            .done(showComparedImagesCore)
            .fail(function() {
                showError((lng !== 'ja') ? 'Image could not be loaded.' : '画像を読み込めませんでした。');
            })
            .always(function() {
                if ($.fn.spin) {
                    $('body').spin(false);
                }
            });
    }
    function init() {
        var params = {};
        if ($.isPlainObject(config)) {
            if ($.isPlainObject(config.data)) {
                params = config.data;
            } else {
                params = getParams(location.search) || {};
            }
        } else {
            config = {};
        }
        if (config.id) {
            var i;
            var sequence = params.sequence;
            if (!sequence) {
                var imgs = [];
                var labels = [];
                for (i = 0; i < 100; i++) {
                    var imgvar = 'img' + i.toString();
                    if (params[imgvar]) {
                        imgs.push(params[imgvar]);
                        labels.push(params['img' + i.toString() + '_label']);
                    }
                }
                sequence = [];
                for (i = 1; i < imgs.length; i++) {
                    var oneData = {
                        img1: imgs[i - 1],
                        img2: imgs[i]
                    };
                    if (labels[i - 1]) {
                        oneData['img1_label'] = labels[i - 1];
                    }
                    if (labels[i]) {
                        oneData['img2_label'] = labels[i];
                    }
                    sequence.push(oneData);
                }
            } else {
                var sequence_ = [];
                for (i = 0; i < sequence.length; i++) {
                    if (sequence[i].img1 && sequence[i].img2) {
                        sequence_.push(sequence[i]);
                    }
                }
                sequence = sequence_;
            }
            if (sequence.length > 1) {
                params.sequence = sequence;
                showComparedImages(config, params);
            } else {
                showError((lng !== 'ja') ? 'Specify the images to compare.' : '比較する画像を指定してください。');
            }
        } else {
            showError((lng !== 'ja') ? 'Specify where to display.' : '表示先を指定してください。');
        }
    }

    if (typeof cv !== 'undefined' && cv.runtimeInitialized) {
        init();
    } else {
        $(window).on('opencvRuntimeInitialized', init);
    }

    return {
        getParameters: function() {
            return (exportFunctions) ? exportFunctions.getParameters() : {};
        },
        getSequence: function() {
            return (exportFunctions) ? exportFunctions.getSequence() : [];
        }
    };
};